/*
 * This source code is a C++ implementation of task swapping graphs.
 * Copyright (C) 2011, Dohan Kim <dkim@airesearch.kr>
 * Author : Dohan Kim < dkim@airesearch.kr >
 *
 * File Name : PGroup.cpp
 * Description : Main Library of (finite) Permutation Groups
 */

#include "PGroup.h"
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <algorithm>
#include <fstream>
#include <iostream>
#include <math.h>

using namespace std;

PGroup::PGroup()
{
}

PGroup::~PGroup()
{
}

int *PGroup::inverse( int *perm, int length )
{
	int *inverse_perm = new int[length];

	for( int i = 0; i < length; i++ )
	{
		inverse_perm[perm[i] - 1] = i + 1;		
	}

	return inverse_perm;
}

int *PGroup::multiplication( int *perm1, int *perm2, int length )
{
	int *return_perm = new int[length];

	for( int i = 0; i < length; i++ )
	{
		return_perm[i] = perm1[perm2[i] - 1];
	}

	return return_perm;
}

vector< vector<int> > PGroup::disjoint_cycles( int *perm, int length )
{
	vector< vector<int> > d_cycle;
	vector<bool> mark;
	int start = 0;

	mark.resize( length );

	for( int k = 0; k < length; k++ )
	{
		mark[k] = false;
	}

	for( int i = 0; i < length; i++ )
	{
		if( mark[i] == true )
			continue;

		vector<int> temp;

		if( perm[i] == i + 1 )
		{
			mark[i] = true;
			temp.push_back( i + 1 );
			d_cycle.push_back( temp );
		}

		else
		{
			start = i;

			do {
				mark[start] = true;
				temp.push_back( start + 1 );
				start = perm[start] - 1;
			} while ( !mark[start]);

			d_cycle.push_back( temp );
		}

		temp.clear();
	}


	return d_cycle;
}

vector< vector<int> > PGroup::canonical_disjoint_cycles( int bipartite_index, vector<char> &cycle_type, vector< vector<int> > &d_cycle )
{
#ifdef _VERBOSE
	cout<<"d cycle size:"<<d_cycle.size()<<"index:"<<bipartite_index<<endl;
#endif

	bool low_cnt, high_cnt;

	cycle_type.resize( d_cycle.size() );

	for( int i = 0; i < (int)d_cycle.size(); i++ )
	{
		low_cnt = false;
		high_cnt = false;

		for( int j = 0; j < (int)d_cycle[i].size(); j++ )
		{
			if( d_cycle[i][j] <= bipartite_index )
			{
				low_cnt = true;
			}
			if( d_cycle[i][j] > bipartite_index )
			{
				high_cnt = true;
			}
		}
		if( low_cnt && high_cnt )
		{
			cycle_type[i] = 'S'; //Simple Mixed type
		}
		else if( low_cnt && !high_cnt && d_cycle[i].size() > 1 )
		{
			cycle_type[i] = 'I'; //Internal
		}
		else if( !low_cnt && high_cnt && d_cycle[i].size() > 1 )
		{
			cycle_type[i] = 'E'; //External
		}	
		else
		{
			cycle_type[i] = 'N'; // Cycle of length 1
		}
	} 

	for( int i = 0; i < (int)d_cycle.size(); i++ )
	{
		int temp = 0;
		vector<int>::iterator it;

		if( cycle_type[i] == 'S' )
		{
			while( d_cycle[i].back() <= bipartite_index )
			{
				temp = d_cycle[i].back();
				d_cycle[i].pop_back();
				
				d_cycle[i].insert( d_cycle[i].begin(), temp );	
			} 
			while( d_cycle[i].front() > bipartite_index )
			{
				temp = d_cycle[i].front();
				d_cycle[i].erase( d_cycle[i].begin(), d_cycle[i].begin() + 1 );
				
				d_cycle[i].insert( d_cycle[i].end(), temp );	
			} 
		}
	}

	for( int i = 0; i < (int)d_cycle.size(); i++ )
	{
		high_cnt = false;

		if( cycle_type[i] == 'S' )
		{
			for( int j = 0; j < (int)d_cycle[i].size(); j++ )
			{
				if( d_cycle[i][j] > bipartite_index )
				{
					high_cnt = true;
				}
				if( d_cycle[i][j] <= bipartite_index )
				{
					if( high_cnt )
					{
						cycle_type[i] = 'M'; //mixed type
						break;
					}	
				}
			}
		}
	}

#ifdef _VERBOSE
	for( int k = 0; k < (int) cycle_type.size(); k++ )
	{
		cout<<"Cycle Type:"<<cycle_type[k]<<endl;
	}

	for( int i = 0; i < (int) d_cycle.size(); i++ )
	{
		for( int j = 0; j < (int) d_cycle[i].size(); j++ )
		{
			cout<<d_cycle[i][j]<<" ";
		}
		cout<<endl;
	}
#endif

	return d_cycle;
}

int PGroup::bipartite_internal_cycle_factorization( int bipartite_index,
        vector<transposition> &complete_bipartite_sequence, vector<int> &cycle )
{
	/* You can replace the below "bipartite_index+1  with any number in [bipartite_index+1,n],*/
	/*  where n is the number of tasks (agents). */
	complete_bipartite_sequence.push_back( transposition( cycle[0], bipartite_index + 1 ) );

	for( int i = (int)cycle.size() - 1; i >= 0; i-- )
	{
		complete_bipartite_sequence.push_back( transposition( cycle[i], bipartite_index + 1 ) );
	}

	return 1;
}
int PGroup::bipartite_external_cycle_factorization( int bipartite_index,
        vector<transposition> &complete_bipartite_sequence, vector<int> &cycle )
{
	for( int i = (int)cycle.size() - 1; i >= 0; i-- )
	{
		/* You can replace the below 1 with any number in [1,k], where k is a bipartite index. */
		complete_bipartite_sequence.push_back( transposition( 1, cycle[i] ) );
	}

	complete_bipartite_sequence.push_back( transposition( 1, cycle.back() ) );

	return 1;
}
int PGroup::bipartite_simple_cycle_factorization( int bipartite_index,
        vector<transposition> &complete_bipartite_sequence, vector<int> &cycle )
{
	int low_cnt = 0, high_cnt = 0;

	for( int i = 0; i < (int) cycle.size(); i++ )
	{
		if( cycle[i] <= bipartite_index )
		{
			low_cnt++;
		}
		else
		{
			high_cnt++;
		}
	}

	if( low_cnt == 1 && high_cnt == 1 )
	{
		complete_bipartite_sequence.push_back( transposition( cycle[0], cycle[1] ) );
	}
	else if( low_cnt == 1 && high_cnt > 1 )
	{
		for( int j = (int) cycle.size() - 1; j > 0; j-- )
		{
			complete_bipartite_sequence.push_back( transposition( cycle[0], cycle[j] ) );
		}
	}
	else if( low_cnt > 1 && high_cnt == 1)
	{
		for( int k = (int) cycle.size() - 2; k >= 0; k-- )
		{
			complete_bipartite_sequence.push_back( transposition( cycle[k], cycle[(int)cycle.size()-1] ) );
		}
	}
	else if( low_cnt > 1 && high_cnt > 1 )
	{
		int bipartite_anchor_index = 0;

		for( int m = 0; m < (int) cycle.size(); m++ )
		{
			if( cycle[m] > bipartite_index )
			{
				bipartite_anchor_index = m;
				break;
			}
		}

#ifdef _VERBOSE
		cout<<"anchor index:"<<bipartite_anchor_index<<endl;
#endif

		for( int s = (int)cycle.size() - 1; s > bipartite_anchor_index; s-- )
		{
			complete_bipartite_sequence.push_back( transposition( cycle[0], cycle[s] ) );
		}

		for( int t = bipartite_anchor_index - 1; t >= 0; t-- )
		{
			complete_bipartite_sequence.push_back( transposition( cycle[t], cycle[bipartite_anchor_index] ) );
		}
	}
	else
	{
		cout<<"Cannot be reached here:"<<__FILE__<<__LINE__<<endl;
		exit( -1 );
	}

	return 1;
}

int PGroup::bipartite_mixed_cycle_factorization( int bipartite_index,
        vector<transposition> &complete_bipartite_sequence, vector<int> &cycle )
{
	bool low_cnt = false, high_cnt = false, high_cnt_for_type_check = false, mixed_type = false;
	int split_index = 0;

	/* Copying vector contents for recursive operations */
	vector<int> simple_cycle( cycle );
	vector<int> remainder_mixed_cycle( cycle );

	for( int i = 0; i < (int) cycle.size(); i++ )
	{
		if( cycle[i] <= bipartite_index )
		{
			if( !high_cnt )
			{
				low_cnt = true;
			}
			else
			{
				split_index = i - 1;	
				break;
			}
		}
		else
		{
			high_cnt = true;
		}
	}

#ifdef _VERBOSE
	cout<<"Split index:"<<split_index<<endl;
#endif

	simple_cycle.erase( simple_cycle.begin() + split_index + 1, simple_cycle.end() );
	remainder_mixed_cycle.erase( remainder_mixed_cycle.begin(), remainder_mixed_cycle.begin() + split_index + 1 );

#ifdef _VERBOSE
	cout<<"Remainder:";
	for( int j = 0; j < (int) remainder_mixed_cycle.size(); j++ )
	{
		cout<<remainder_mixed_cycle[j]<<" ";
	}

	cout<<"Simple cycle:";
	for( int k = 0; k < (int) simple_cycle.size(); k++ )
	{
		cout<<simple_cycle[k]<<" ";
	}
#endif

	/* The following describes the recursive factorization of a mixed cycel. */ 
	/* See Algorithm 5 in the reference. */

	int index = complete_bipartite_sequence.size();

	this->bipartite_simple_cycle_factorization( bipartite_index,
        complete_bipartite_sequence, simple_cycle );

	if( complete_bipartite_sequence[index].first == simple_cycle[0] )  
	{
		complete_bipartite_sequence.insert( complete_bipartite_sequence.begin() + index + 1, 
			transposition( remainder_mixed_cycle[0], complete_bipartite_sequence[index].second ) );
	}
	else
	{
		complete_bipartite_sequence.insert( complete_bipartite_sequence.begin() + index + 1, 
			transposition( remainder_mixed_cycle[0], complete_bipartite_sequence[index].second ) );
		complete_bipartite_sequence.insert( complete_bipartite_sequence.begin() + index + 2, 
			transposition( simple_cycle[0], complete_bipartite_sequence[index].second ) );
		complete_bipartite_sequence.insert( complete_bipartite_sequence.begin() + index + 3, 
			transposition( remainder_mixed_cycle[0], complete_bipartite_sequence[index].second ) );
	}

	for( int m = 0; m < (int)remainder_mixed_cycle.size(); m++ )
	{
		if( remainder_mixed_cycle[m] > bipartite_index )
		{
			high_cnt_for_type_check = true;
		}
		if( remainder_mixed_cycle[m] <= bipartite_index )
		{
			if( high_cnt_for_type_check )
			{
				mixed_type = true;
				break;
			}
		}
	}

	if( mixed_type )
	{
		this->bipartite_mixed_cycle_factorization( bipartite_index,
			complete_bipartite_sequence, remainder_mixed_cycle );
	}
	else
	{
		this->bipartite_simple_cycle_factorization( bipartite_index,
			complete_bipartite_sequence, remainder_mixed_cycle );
	}

	return 1;
}

int PGroup::line_inversion_number( int *perm, int length )
{
	int inv_number = 0;

	for( int i = 0; i < length; i++ )
	{
		for( int j = i + 1; j < length; j++ )
		{
			if( perm[i] > perm[j] )
			{
				inv_number++;
			}
		}
	}

	return inv_number;
}

int PGroup::ring_inversion_number( int *dvec, int length )	
{
	int inv_number = 0;

	for( int i = 0; i < length; i++ )
	{
		for( int j = i + 1; j < length; j++ )
		{
			if( i + 1 + dvec[i] > j + 1 + dvec[j] ||  i + 1 + dvec[i] + length < j + 1 + dvec[j] )
			{
				inv_number++;
			}
		}
	}

	return inv_number;
}

bool compare( int i, int j ) 
{ 
	return i<j; 
}

int *PGroup::stable_displacement_vector( int *perm, int length )
{
	int *dvec = new int[length];
	int *max, *min = NULL;

	for( int i = 0; i < length; i++ )
	{
		dvec[i] = perm_element_find( perm, length, i + 1 ) - ( i + 1 ); 
	}

	max = max_element( dvec, dvec + length, compare );
	min = min_element( dvec, dvec + length, compare );

	while( *max - *min > length )
	{
		*max = *max - length;
		*min = *min + length;

		max = max_element( dvec, dvec + length, compare );
		min = min_element( dvec, dvec + length, compare );
	}
	
	return dvec;
}

int PGroup::perm_element_find( int *perm, int length, int k )
{
	for( int i = 0; i < length; i++ )
	{
		if( perm[i] == k )
		{
			return i + 1;
		}
	}

	return 1;
}
