/*
 * This source code is a C++ implementation of task swapping graphs.
 * Copyright (C) 2011, Dohan Kim <dkim@airesearch.kr>
 * Author : Dohan Kim < dkim@airesearch.kr >
 *
 *
 * File Name : TSG.cpp
 * Description : Main Library of TSG
 *
 */

#include "TSG.h"
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <algorithm>
#include <fstream>
#include <iostream>
#include <math.h>

using namespace std;

TSG::TSG( TSGIO *tsgIO, PGroup *pgroup )
{
	this->tsgIO = tsgIO;
	this->pgroup = pgroup;
}

vector<int> TSG::perm_bubble_sorting( int *perm, int length )
{
    vector<int> generator_sequence;
	int *temp = new int[length];

	for( int k = 0; k < length; k++ )
	{
		temp[k] = perm[k];
	}

	for( int i = length - 1; i > 0; i-- )
	{
		for( int j = 0; j < i; j++ )
		{
			if( temp[j] > temp[j + 1] )
			{
				generator_sequence.push_back( j );
				swap(temp[j], temp[j+1]);
			}
		}
	}

    return generator_sequence;
}

int TSG::tsg_print_result( vector<transposition> &minimum_length_sequence, int upper_bound )
{
	cout<<endl<<"Minimum number of adjacent task swappings needed from a source task assignment to target task assignment: ";
	cout<<minimum_length_sequence.size()<<endl;

	cout<<"Minimum length-sequence of adjacent task swappings: "<<endl<<"(";
	for( int j = 0; j < (int)minimum_length_sequence.size() ; j++ )
	{
		if( j != (int)minimum_length_sequence.size() - 1 )
		{
			cout<<"("<<minimum_length_sequence[j].first<<" "<<minimum_length_sequence[j].second<<"),";

			if( j % 50 == 0 && j != 0 )
			{
				cout<<endl;
			}
		}
		else
		{
			cout<<"("<<minimum_length_sequence[j].first<<" "<<minimum_length_sequence[j].second<<")";
		}
	}
	cout<<")"<<endl<<endl;

#ifdef _DEBUG
	int *src_task_assignment = new int[tsgIO->number_of_tasks];
	int *dst_task_assignment = new int[tsgIO->number_of_tasks];

	for( int n = 0; n < tsgIO->number_of_tasks; n++ )
	{
		src_task_assignment[n] = tsgIO->source_task_assignment[n];
		dst_task_assignment[n] = tsgIO->target_task_assignment[n];
	}
	cout<<"Source Task Assignment: ";
	for( int s = 0 ; s < tsgIO->number_of_tasks ; s++ )
	{
		cout<<src_task_assignment[s]<<" ";
	}
	cout<<endl;


	for( int k = 0; k < (int)minimum_length_sequence.size() ; k++ )
	{
		cout<<k+1<<". "<<"After swapping "<<minimum_length_sequence[k].first<<" and "<<minimum_length_sequence[k].second<<" positions: ";
		swap( src_task_assignment[minimum_length_sequence[k].first - 1], src_task_assignment[minimum_length_sequence[k].second - 1] );

		for( int v = 0 ; v < tsgIO->number_of_tasks ; v++ )
		{
			cout<<src_task_assignment[v]<<" ";
		}
		cout<<endl;
	}
		
	cout<<"Target Task Assignment: ";
	for( int t = 0 ; t < tsgIO->number_of_tasks ; t++ )
	{
		cout<<dst_task_assignment[t]<<" ";
	}
	cout<<endl;

	bool compare_assignments = true;

	for( int u = 0 ; u < tsgIO->number_of_tasks ; u++ )
	{
		if( src_task_assignment[u] != dst_task_assignment[u] )
		{
			compare_assignments = false;
			break;
		}
	}

	cout<<endl<<"Upper bound: "<<upper_bound<<endl;

	if( compare_assignments )
	{
		cout<<"Reached the target task assignment successfully"<<" ("<<(int)minimum_length_sequence.size()<<" steps)."<<endl;
	}
	else
	{
		cout<<"Not reached the target task assignment."<<endl;
	}
#endif

	return 1;
}

int TSG::tsg_line()
{
	int line_upper_bound = 0;

	vector<transposition> line_sequence;

	int *temp = 
		pgroup->multiplication( pgroup->inverse( tsgIO->target_task_assignment, tsgIO->number_of_tasks ), 
		tsgIO->source_task_assignment, tsgIO->number_of_tasks );

	vector<int> l_sequence =
		perm_bubble_sorting( temp, tsgIO->number_of_tasks );

	for( int i = 0; i < (int)l_sequence.size(); i++ )
	{
		line_sequence.push_back( transposition( l_sequence[i] + 1, l_sequence[i] + 2 ) );
	}

#ifdef _DEBUG
	line_upper_bound = ( tsgIO->number_of_tasks * ( tsgIO->number_of_tasks - 1 ) ) / 2;
	cout<<endl<<"Upper bound (diameter of bubble sort Cayley graph BS_n for n="<<tsgIO->number_of_tasks<<") : ";
	cout<<line_upper_bound<<" (i.e.,";
	cout<<"nx(n-1)/2 for n="<<tsgIO->number_of_tasks <<")."<<endl;
#endif
	this->tsg_print_result( line_sequence, line_upper_bound );

	return 1;
}

int TSG::tsg_star()
{
	int star_upper_bound = 0;

	vector<transposition> star_sequence;

	int *temp =
		pgroup->multiplication( pgroup->inverse( tsgIO->source_task_assignment, tsgIO->number_of_tasks ),
		tsgIO->target_task_assignment, tsgIO->number_of_tasks );

    vector<int> s_sequence;
	vector< vector<int> > d_cycle = pgroup->disjoint_cycles( temp, tsgIO->number_of_tasks );

	for( int i = 0; i < (int) d_cycle.size(); i++ )
	{
		if( d_cycle[i].size() > 1 )
		{
			if( d_cycle[i][0] == 1 )
			{
				for( int j = d_cycle[i].size() - 1; j > 0; j-- )
				{
					s_sequence.push_back( d_cycle[i][j] );
				}
			}
			else
			{
				s_sequence.push_back( d_cycle[i][0] );

				for( int k = d_cycle[i].size() - 1; k >= 0; k-- )
				{
					s_sequence.push_back( d_cycle[i][k] );
				}
			}
		}
	}

	for( int m = 0; m < (int)s_sequence.size(); m++ )
	{
		star_sequence.push_back( transposition( 1, s_sequence[m] ) );
	}

#ifdef _DEBUG
	star_upper_bound = (int)floor(  3 * ( tsgIO->number_of_tasks - 1 ) / 2 );
	cout<<endl<<"Upper bound (diameter of star Cayley graph ST_n for n="<<tsgIO->number_of_tasks<<") : ";
	cout<<star_upper_bound<<" (i.e.,";
	cout<<"floor(3*(n-1)/2) for n="<<tsgIO->number_of_tasks <<")."<<endl;
#endif

	this->tsg_print_result( star_sequence, star_upper_bound );
	
	return 1;
}

int TSG::tsg_complete()
{
	int complete_upper_bound = 0;

	vector<transposition> complete_sequence;

	int *temp =
		pgroup->multiplication( pgroup->inverse( tsgIO->source_task_assignment, tsgIO->number_of_tasks ),
		tsgIO->target_task_assignment, tsgIO->number_of_tasks );

	vector< vector<int> > c_sequence;
	vector< vector<int> > d_cycle = pgroup->disjoint_cycles( temp, tsgIO->number_of_tasks );

	for( int i = 0; i < (int) d_cycle.size(); i++ )
	{
		if( d_cycle[i].size() > 1 )
		{
			vector<int> temp;
			temp.push_back( d_cycle[i][0] );

			for( int j = d_cycle[i].size() - 1; j > 0; j-- )
			{
				temp.push_back( d_cycle[i][j] );
			}
			c_sequence.push_back( temp );
			temp.clear();
		}
	}

	for( int m = 0; m < (int)c_sequence.size(); m++)
	{
		for( int n = 1; n < (int)c_sequence[m].size(); n++ )
		{
			complete_sequence.push_back( transposition( c_sequence[m][0], c_sequence[m][n] ) );
		}
		cout<<endl;
	}

#ifdef _DEBUG
	complete_upper_bound =  tsgIO->number_of_tasks - 1;
	cout<<endl<<"Upper bound (diameter of complete transposition graph CT_n for n="<<tsgIO->number_of_tasks<<") : ";
	cout<<complete_upper_bound<<" (i.e.,";
	cout<<"n-1 for n="<<tsgIO->number_of_tasks <<")."<<endl;
#endif

	this->tsg_print_result( complete_sequence, complete_upper_bound );

	return 1;
}

int TSG::tsg_complete_bipartite()
{
	int complete_bipartite_upper_bound = 0;

	vector<transposition> complete_bipartite_sequence;

	int *temp =
		pgroup->multiplication( pgroup->inverse( tsgIO->source_task_assignment, tsgIO->number_of_tasks ),
		tsgIO->target_task_assignment, tsgIO->number_of_tasks );

	vector< vector<int> > d_cycle = pgroup->disjoint_cycles( temp, tsgIO->number_of_tasks );
	vector<char> cycle_type;

#ifdef _VERBOSE
	cout<< "Disjoint Cycle(s):"<<endl;

	for( int i = 0; i < (int) d_cycle.size(); i++ )
	{
		for( int j = 0; j < (int) d_cycle[i].size(); j++ )
		{
			cout<<d_cycle[i][j]<<" ";
		}
		cout<<endl;
	}
#endif

	pgroup->canonical_disjoint_cycles( tsgIO->bipartite_index, cycle_type, d_cycle );

	for( int i = 0; i < (int) d_cycle.size(); i++ )
	{
		if( cycle_type[i] == 'I' )
		{
			pgroup->bipartite_internal_cycle_factorization( tsgIO->bipartite_index, 
				complete_bipartite_sequence, d_cycle[i] ); 
		}
		else if( cycle_type[i] == 'E' )
		{
			pgroup->bipartite_external_cycle_factorization( tsgIO->bipartite_index, 
				complete_bipartite_sequence, d_cycle[i] ); 
		}
		else if( cycle_type[i] == 'S' )
		{
			pgroup->bipartite_simple_cycle_factorization( tsgIO->bipartite_index, 
				complete_bipartite_sequence, d_cycle[i] ); 
		}
		else if( cycle_type[i] == 'M' )
		{
			pgroup->bipartite_mixed_cycle_factorization( tsgIO->bipartite_index, 
				complete_bipartite_sequence, d_cycle[i] ); 
		}
		else /* cycle of length 1 */
		{
			continue;
		}
	}

#ifdef _DEBUG
	complete_bipartite_upper_bound =  
		tsgIO->number_of_tasks - 1 + 
			max( (int) floor( tsgIO->bipartite_index / 2 ), 
				(int) floor( ( tsgIO->number_of_tasks - tsgIO->bipartite_index ) / 2 ) ) ;
	cout<<endl<<"Upper bound (diameter of generalized star graph GST_{n,k} for n=";
	cout<<tsgIO->number_of_tasks<<", k="<<tsgIO->bipartite_index<<"): ";
	cout<<complete_bipartite_upper_bound<<" (i.e.,";
	cout<<"n - 1 + max( floor( k / 2 ), floor( ( n - k ) / 2 ) )";
 	cout<<" for n="<<tsgIO->number_of_tasks<<", k="<<tsgIO->bipartite_index<<")."<<endl;
#endif

	this->tsg_print_result( complete_bipartite_sequence, complete_bipartite_upper_bound );

	return 1;
}

int TSG::tsg_ring()
{
	int ring_upper_bound = 0;

	int *perm_for_sorting =
		pgroup->multiplication( pgroup->inverse( tsgIO->target_task_assignment, tsgIO->number_of_tasks ),
		tsgIO->source_task_assignment, tsgIO->number_of_tasks );

	int *d_vec = pgroup->stable_displacement_vector( perm_for_sorting, tsgIO->number_of_tasks ); 

	vector<int> r_sequence;
	vector<transposition> ring_sequence;

#ifdef _VERBOSE
	cout<<"Stable displacement vector : ";

	for( int i = 0; i < tsgIO->number_of_tasks; i++)
	{
		cout<<d_vec[i]<<" ";
	}
	cout<<endl;
#endif

	int i_number = pgroup->ring_inversion_number( d_vec, tsgIO->number_of_tasks );

#ifdef _VERBOSE
	int	temp = i_number;
#endif

#ifdef _DEBUG
	cout<<"Inversion number (for p_2^{-1}p_1, where p_1, p_2 denote permutations for source and target task assignment, respectively): "<<i_number<<endl;
#endif

	for( int j = i_number; j > 0 ; j-- ) 
	{
		perm_ring_sorting( perm_for_sorting, d_vec, tsgIO->number_of_tasks, r_sequence );
		d_vec = pgroup->stable_displacement_vector( perm_for_sorting, tsgIO->number_of_tasks );

#ifdef _VERBOSE
		cout<<temp - j + 1<<"." "Stable displacement vector "<<"(inversion #:"<<i_number<<")  ";
		for( int i = 0; i < tsgIO->number_of_tasks; i++)
    	{
			cout<<d_vec[i]<<" ";
		}
		cout<<endl;
#endif
		i_number = pgroup->ring_inversion_number( d_vec, tsgIO->number_of_tasks );

#ifdef _VERBOSE
	cout<<"Inversion number (Ring Topology): "<<i_number<<endl;
#endif
	}		

#ifdef _VERBOSE
	for( int m = 0; m < (int)r_sequence.size(); m++)
	{
		cout<<r_sequence[m]<<" ";
	}

	cout<<"Perm for sorting:";

	for( int n = 0; n < tsgIO->number_of_tasks; n++)
	{
		cout<<perm_for_sorting[n]<<" ";
	}
#endif

	for( int v = 0; v < (int)r_sequence.size(); v++ )
	{
		if( r_sequence[v] == tsgIO->number_of_tasks )
		{
			ring_sequence.push_back( transposition( 1, r_sequence[v] ) );
		}
		else
		{
			ring_sequence.push_back( transposition( r_sequence[v], r_sequence[v] + 1  ) );
		}
	}

#ifdef _DEBUG
	ring_upper_bound = ( tsgIO->number_of_tasks *  tsgIO->number_of_tasks ) / 4;
	cout<<endl<<"Conjectured upper bound (diameter of modified bubble sort Cayley graph MBS_n for n="<<tsgIO->number_of_tasks<<") : ";
	cout<<ring_upper_bound<<" (i.e.,";
	cout<<"n^2/4 for n="<<tsgIO->number_of_tasks <<")."<<endl;
#endif

	this->tsg_print_result( ring_sequence, ring_upper_bound );

	cout<<endl<<endl;

	return 1;
}

int TSG::perm_ring_sorting( int *perm_for_sorting, int *d_vec, int number_of_tasks, vector<int> &r_sequence )
{
	for( int i = 0; i < number_of_tasks; i++ )
	{
		if( i != number_of_tasks - 1 )
		{
			if( d_vec[perm_for_sorting[i]-1] < 0 )
			{
				if( d_vec[perm_for_sorting[i+1]-1] > 0 )
				{
					r_sequence.push_back( i + 1 );
#ifdef _VERBOSE
					cout<<"Before: "<<perm_for_sorting[i]<<endl;
					swap( perm_for_sorting[i], perm_for_sorting[i+1] );
					cout<<"After: "<<perm_for_sorting[i]<<endl;
#else
					swap( perm_for_sorting[i], perm_for_sorting[i+1] );
#endif

					break;
				}
			}
			else if( d_vec[perm_for_sorting[i]-1] == 0 )
			{	
				if( d_vec[perm_for_sorting[i+1]-1] > 1 )
				{
					r_sequence.push_back( i + 1 );
					swap( perm_for_sorting[i], perm_for_sorting[i+1] );
					break;
				}
			}
			else
			{
				continue;
			}
		}
		else // compare the first element because it is a ring topology
		{	
			if( d_vec[perm_for_sorting[i]-1] < 0 )
			{
				if( d_vec[perm_for_sorting[0]-1] > 0 )
				{
					r_sequence.push_back( i + 1 );
					swap( perm_for_sorting[i], perm_for_sorting[0] );
					break;
				}
			}
            else if( d_vec[perm_for_sorting[i]-1] == 0 )
            {   
                if( d_vec[perm_for_sorting[0]-1] > 1 )
                {
                    r_sequence.push_back( i + 1 );
                    swap( perm_for_sorting[i], perm_for_sorting[0] );
                    break;
                }
            }
			else
				continue;
		}

#ifdef _VERBOSE
		for( int n = 0; n < tsgIO->number_of_tasks; n++)
		{
			cout<<perm_for_sorting[n]<<" ";
		}
		cout<<endl;
#endif
	}

	return 1;
}

int TSG::print()
{
#ifdef _VERBOSE	
	cout<<"Inverse (source): ";
	int *temp1 = pgroup->inverse( tsgIO->source_task_assignment, tsgIO->number_of_tasks );

    for( int i = 0; i < tsgIO->number_of_tasks; i++ )
    {
        cout<<temp1[i]<<" ";
    }
	
	cout << endl;

	cout<<"Multiplication (source, target): ";

	int *temp2 = pgroup->multiplication( tsgIO->source_task_assignment, tsgIO->target_task_assignment, tsgIO->number_of_tasks );

	for( int i = 0; i < tsgIO->number_of_tasks; i++ )
	{
		cout<<temp2[i]<<" ";
	}

	cout<<endl;

	cout << "Inversion number (source): ";
	int inv = pgroup->line_inversion_number( tsgIO->source_task_assignment, tsgIO->number_of_tasks );

	cout << inv <<endl;

	vector< vector <int> > dcycle = 
		pgroup->disjoint_cycles ( tsgIO->target_task_assignment, tsgIO->number_of_tasks );
	cout<< "Disjoint Cycle(s):"<<endl;

	for( int i = 0; i < (int) dcycle.size(); i++ )
	{
		for( int j = 0; j < (int) dcycle[i].size(); j++ )
		{
			cout<<dcycle[i][j];
		}
		cout<<endl;
	}

	cout<<endl;

	cout<< "Permutation Bubble Sorting:";

	vector<int> bsequence = 
		perm_bubble_sorting( tsgIO->source_task_assignment, tsgIO->number_of_tasks );

	for( int i = 0; i < (int)bsequence.size(); i++)
	{
		cout<<bsequence[i]<<" ";
	}

	cout<<endl;

	int test_vec[8] = {7, 8, 4, 6, 3, 5, 1, 2};

	int *d_vec = pgroup->stable_displacement_vector( test_vec, 8 ); 

	for( int i = 0; i < tsgIO->number_of_tasks; i++)
	{
		cout<<d_vec[i]<<" ";
	}

	int i_number = pgroup->ring_inversion_number( d_vec, 8 );

	cout<<"Ring inv number: "<<i_number<<endl;

	cout<<endl;
#endif

	return 1;
}
